--[[
    Knowing this helped me a lot : https://stackoverflow.com/questions/4911186/difference-between-and-in-lua
    " The colon is for implementing methods that pass self as the first parameter. So x:bar(3,4)should be the same as x.bar(x,3,4)."  
]]

TheMotherTongue = {} -- global table to access the addon
local TMT = TheMotherTongue -- local name for ease of work
TMT.version = "v1.14.3_2.0"
--[[
 Global variables to allow other addons to change the names used if they want to use another language.
 It has to be collections of the form : {{id, name}, {id, name}, ...}
]] 
TMT.npcs = {}
TMT.spells = {}
TMT.items = {}
TMT.quests = {}
TMT.talents = {}
TMT.areas = {} -- Gotten from C_Map.GetAreaInfo(areaID)

--TMT.instances = {} -- Gotten from GetInstanceInfo
--TMT.maps = {} -- Gotten from C_Map.GetBestMapForUnit("player") or WorldMapFrame:GetMapID()

--[[
    A word on colors in TMT :
    To display text in color in Blizzard UI easily, we have to use the hex form,
    preceded by "|cFF".
    User colors preferences are stored in RGB. It's because the color picker works
    in rgb and not in hexadecimal.
    Oh, and rgb in color picker is between 0 and 1.
]]
local TMT_Colors = {
    orange = "|cFFFF8000",
    red = "|cFFFF3333",
    blue = "|cFF0080FF",
    magenta = "|cFFFF0040",
    grey = "|cFF191203",
    lightGrey = "|cFFB3B3B3",
    brown = "|cFF5E130B",
    lightBrown = "|cFFAC240B",
    green = "|cFF5ABE29",
    white = "|cFFFFFFFF",
    classes = {
        WARRIOR = "|cFFC79C6E",
        PALADIN = "|cFFF58CBA",
        HUNTER = "|cFFA9D271",
        ROGUE = "|cFFFFF569",
        PRIEST = "|cFFFFFFFF",
        DEATHKNIGHT = "|cFFC41F3B",
        SHAMAN = "|cFF0070DE",
        MAGE = "|cFF40C7EB",
        WARLOCK = "|cFF8787ED",
        MONK = "|cFF00FF96",
        DRUID = "|cFFFF7D0A",
        DEMONHUNTER = "|cFFA330C9"
    }
}

local TMT_Languages = {
    en = "English",
    fr = "Français",
    de = "Deutsch",
    es = "Español",
    pt = "Português",
    it = "Italiano",
    ru = "pусский",
    ko = "한국어",
    cn = "简体中文",
}

-- default values of settings that the user can change
local TMT_Defaults = {
    language = "",
    npcColor = {255, 128, 0}, -- name color in tooltip
    questColor = {255, 106, 0}, -- names color when next to the quests log
    spellColor = {0, 128, 255}, -- name color in tooltip
    questsFrame_HighlightBackdropColor = {255, 94, 80, 0.704594045877457}, -- background color of the selected quest
    itemColor = {87, 152, 255}, -- name color in tooltip
    displayNPCsNames = true,
    npcTooltip_displayNameInTitle = false,
    displayQuestsNames = true,
    displaySpellsNames = true,
    questsNamesToggle = true, -- remembers if the quests frame is to be expanded or not
    spellTooltip_displayNameInTitle = false,
    displayItemsNames = true,
    itemTooltip_displayNameInTitle = false,
    itemTooltip_UseBlizzardsColor = false,
    questsNames_UseBlizzardsColor = false,
    questsNames_DisplayQuestsTags = true,
    questsFrameX = -39.5,
    questsFrameY = -70,
    questsFrame_PositionChangedByUser = false,
    questsFramePoint = "TOPLEFT",
    questsFrameRelativePoint = "TOPRIGHT",
    spellTooltip_UseClassColor = false,
    tooltips_DisplayNameUnderTitle = false,
    displayAreasNames = true,
    areaTooltip_DisplayNameInTitle = true,
    lastUpdateMessageVersion = "",
}

local TMT_AvailableLanguages = {}
local TMT_VIncrementsPerChar = 12.5 -- used to "correctly" resize quests frame according to numbers of entries in the quest log
local TMT_VMargin = 16 -- used for margin when displaying quests names in quests frame
local TMT_HMargin = 15 -- same
local TMT_Tab = "      "
local TMT_QuestsLogEntries = 0 -- used to update the quests frame only if entries collapsing changed
local TMT_QuestsFrameToggleButtonX
local TMT_QuestsFrameToggleButtonY
local TMT_ColorBeingChanged -- used by the color picker to know which color to change
local TMT_QuestsFrame -- the frame containing quests English name
local TMT_QuestsToggle -- the toggle button to show/hide the quests frame (it's kinda bulky...)
local TMT_QuestsReset -- button to reset the frame back to default position
local TMT_QuestsFrameParent -- Host of the quests frame. This is to support other addons like QuestLogEx or QuestGuru
local TMT_QuestsFramePool = {} -- Stores all quests buttons and zones headers that are or have been in the questLog since user log in
local TMT_QuestsLogEntriesPool = {} -- Stores names of quests buttons and zones headers that are currently in the questLog
local TMT_HighlightedEntry -- the entry currently highlighted in TMT questsFrame (a button)
-- https://wow.gamepedia.com/API_GetQuestTagInfo
local TMT_QuestsTags = {{1,"Elite"},{41,"PvP"},{62,"Raid"},{81,"Dungeon"},{83,"Legendary"},{85,"Heroic"},{98,"Scenario"},{102,"Account"},{107,"Leatherworking World Quest"}}
local TMT_QuestsTagsSpacing = 0
--[[
    Button to test for text width more accurately. So instead of test string length,
    we set the button text, and then test the fontString width
]] 
local TMT_DebugButton = CreateFrame("Button", TMT_QuestsFrame_GlobalName)
--local TMT_SpellsMaxId = 32061 -- Classic : 32061
local TMT_AreasMaxId = 50000
local TMT_QuestsMaxId = 15000

-- From the time TMT didn't have an option panel. Keepin them just in case
local TMT_SlashCommands = {
    {"npcsnames[on/off]", "Turn [on/off] NPCs names translation"},
    {"npcname[top/bottom]", "Set NPCs names to display at the [top/bottom] of the tooltip"},
    {"questsnames[on/off]", "Turn [on/off] quests names translation"},
    {"spellsnames[on/off]", "Turn [on/off] spells names translation"},
    {"spellname[top/bottom]", "Set spells names to display at the [top/bottom] of the tooltip"},
    {"version", "Displays addon version"}
}

local function TMT_WriteInChat(something)
    local chatFrame = DEFAULT_CHAT_FRAME.editBox
    ChatEdit_ActivateChat(chatFrame)
    chatFrame:SetText(chatFrame:GetText() .. something)
end

--[[ Gets all spells names with their id
local function TMT_GetSpellsNames()
    TMT_NamesDump = {}
    local pair = {}
    local name, rank, icon, castTime, minRange, maxRange

    for i = 1, TMT_SpellsMaxId do
        name, rank, icon, castTime, minRange, maxRange = GetSpellInfo(i)
        if name then
            pair = {i, name}
            table.insert(TMT_NamesDump, pair)
        end
    end

    print(TMT_Colors.blue .. "Done")
end]]


local function TMT_GetTalents()
    local talents = {}
    for tab = 1, 3 do
        local tabPage = {}

        for talent = 1, 30 do
            talentName = GetTalentInfo(tab, talent)
            if talentName then
                table.insert(tabPage, talentName)
            end
        end

        table.insert(talents, tabPage)
    end

    TMT_NamesDump = talents
    print(TMT_Colors.blue .. "Done")
end

--Gets all areas names with their id
local function TMT_GetAreas()
    TMT_NamesDump = {}
    local pair = {}
    local name

    for i = 1, TMT_AreasMaxId do
        name = C_Map.GetAreaInfo(i)
        if name then
            pair = {i, name}
            table.insert(TMT_NamesDump, pair)
        end
    end

    print(TMT_Colors.blue .. "Done")
end

--Gets all quests names with their id
local function TMT_GetQuests()
    TMT_NamesDump = {}
    local pair = {}
    local name

    for i = 1, TMT_QuestsMaxId do
        name = C_QuestLog.GetQuestInfo(i)
        if name then
            pair = {i, name}
            table.insert(TMT_NamesDump, pair)
        end
    end

    print(TMT_Colors.blue .. "Done")
end

local function TMT_GetStringWidth(text, font)
    if font then
        TMT_DebugButton:SetNormalFontObject(font)
    else
        TMT_DebugButton:SetNormalFontObject("QuestFontNormalSmall")
    end

    TMT_DebugButton:SetText(text)
    return math.floor(TMT_DebugButton:GetFontString():GetStringWidth() + 0.5)
end


local function TMT_LanguageDropdown(settings)
    local values = settings['values'] or {}
    local level = 1
    local changeHandler = settings['changeHandler']
    local dropdown = CreateFrame("Frame", settings['name'], settings['parent'], 'UIDropDownMenuTemplate')

    -- Sets the dropdown width
    local dropdownWidth = 0
    for _, value in pairs(values) do 
        local textWidth = TMT_GetStringWidth(value, "GameFontNormal") + 30
        if textWidth > dropdownWidth then
            dropdownWidth = textWidth
        end
    end

    UIDropDownMenu_SetWidth(dropdown, dropdownWidth)
    UIDropDownMenu_SetText(dropdown, TMT_Languages[TMT_Language])

    UIDropDownMenu_Initialize(dropdown, function(self, level, _)
        local info = UIDropDownMenu_CreateInfo()

        for key, value in pairs(values) do
            info.text = value
            info.checked = key == TMT_Language
            info.menuList = key
            info.hasArrow = false
            info.arg1 = key
            info.func = function(info, arg1, arg2, checked)
                info.checked = true
                UIDropDownMenu_SetText(dropdown, info.value)
                changeHandler(dropdown, info)
            end
            UIDropDownMenu_AddButton(info)
        end
    end)

    return dropdown
end

local function TMT_SetLanguage()
    TMT.npcs = _G["TMT_Npcs_" .. TMT_Language]
    TMT.spells = _G["TMT_Spells_" .. TMT_Language]
    TMT.items = _G["TMT_Items_" .. TMT_Language]
    TMT.quests = _G["TMT_Quests_" .. TMT_Language]
    TMT.talents = _G["TMT_Talents_" .. TMT_Language]
    TMT.areas = _G["TMT_Areas_" .. TMT_Language] 
    --TMT.instances = _G["TMT_Instances_" .. TMT_Language]
    --TMT.maps = _G["TMT_Maps_" .. TMT_Language]

    --removes the other languages tables from memory
    for key,val in pairs(TMT_AvailableLanguages) do
        if key ~= TMT_Language then
            _G["TMT_Npcs_" .. key] = nil
            _G["TMT_Spells_" .. key] = nil
            _G["TMT_Items_" .. key] = nil
            _G["TMT_Quests_" .. key] = nil
            _G["TMT_Talents_" .. key] = nil
            _G["TMT_Areas_" .. key] = nil 
            --_G["TMT_Instances_" .. key] = nil
            --_G["TMT_Maps_" .. key] = nil
        end
    end
end

--[[
    Sometimes the hooked scripts are called more than once by the API. This does happen
    for talents, recipes, formulas, patrons.
    This function is used to check if the name is already in the tooltip to not display it
    more than once.
]]
local function TMT_IsNameAlreadyInTooltip(tooltip, name)
    local found = false
    local tooltipName = tooltip:GetName()

    for line = 1, tooltip:NumLines() do
        if string.match(_G[tooltipName .. "TextLeft" .. line]:GetText(), name) then
            found = true
            break
        end
    end
    
    return found
end

local function TMT_RgbToHex(rgb)
    local hexadecimal = ""
    local digitIndex

    local r, g, b = unpack(rgb)
    local value = r * 65536 + g * 256 + b

    if value == 0 then
        hexadecimal = "000000"
    else     
        while value > 0 do
            digitIndex = math.fmod(value, 16) + 1
            value = math.floor(value / 16)
            hexadecimal = string.sub('0123456789ABCDEF', digitIndex, digitIndex) .. hexadecimal	
        end
    end

    while string.len(hexadecimal) < 6 do
        hexadecimal = "0" .. hexadecimal
    end

	return "|cff" .. hexadecimal
end

local function TMT_GetQuestColor(level)
    local questColor

    if TMT_QuestsNames_UseBlizzardsColor then
        local difficultyColor = GetQuestDifficultyColor(level)
        questColor = TMT_RgbToHex({math.floor((difficultyColor.r * 255) + 0.5), 
                                   math.floor((difficultyColor.g * 255) + 0.5), 
                                   math.floor((difficultyColor.b * 255) + 0.5)})
    else
        questColor = TMT_RgbToHex(TMT_QuestColor)
    end

    return questColor
end

local function TMT_SetButtonState(param, button)
    if param then
        button:Enable()
    else
        button:Disable()
    end
end

-- Set the color of the quest button text and of all of it's tags too
local function TMT_SetButtonTextColor(button, color)
    -- resets the color of the name
    local name = string.sub(button:GetText(), 11) -- removes the previous color
    button:SetText(color .. name)

    -- resets the color of the tag
    local questID = string.match(button:GetName(), "quest: (%d+)_")
    if questID then
        local tag = _G[questID .. "_Tag_GlobalName"]
        if tag then
            local tagName = string.sub(tag:GetText(), 11) -- removes the previous color
            tag:SetText(color .. tagName)
        end
    end
end

-- Set the position of the button and of all of it's tags too
local function TMT_SetButtonPoints(button)
    button:SetPoint("TOPLEFT")
end

local function TMT_ChangeFramePosition(frame, anchor, parent, parentAnchor, x, y)
    frame:ClearAllPoints()
    frame:SetPoint(anchor, parent, parentAnchor, x, y)
end

-- Used to be sure the config panel is always matching saved variables
local function TMT_ConfigPanelSanityCheck()
    if not TMT_NoLanguagePack then
        TMT_NpcsNamesCheckBox_GlobalName:SetChecked(TMT_DisplayNPCsNames)
        TMT_NpcsInTitleCheckBox_GlobalName:SetChecked(TMT_NPCTooltip_DisplayNameInTitle)

        TMT_SpellsNamesCheckBox_GlobalName:SetChecked(TMT_DisplaySpellsNames)
        TMT_SpellsInTitleCheckBox_GlobalName:SetChecked(TMT_SpellTooltip_DisplayNameInTitle)
        TMT_SpellsClassColorCheckBox_GlobalName:SetChecked(TMT_SpellTooltip_UseClassColor)

        TMT_ItemsNamesCheckBox_GlobalName:SetChecked(TMT_DisplayItemsNames)
        TMT_ItemsInTitleCheckBox_GlobalName:SetChecked(TMT_ItemTooltip_DisplayNameInTitle)
        TMT_ItemsBlizzardColorCheckButton_GlobalName:SetChecked(TMT_ItemTooltip_UseBlizzardsColor)

        TMT_QuestsNamesCheckButton_GlobalName:SetChecked(TMT_DisplayQuestsNames)
        TMT_QuestsBlizzardColorCheckButton_GlobalName:SetChecked(TMT_QuestsNames_UseBlizzardsColor)
        TMT_QuestsTagsCheckButton_GlobalName:SetChecked(TMT_QuestsNames_DisplayQuestsTags)

        TMT_AreasNamesCheckBox_GlobalName:SetChecked(TMT_DisplayAreasNames)
        TMT_AreasInTitleCheckBox_GlobalName:SetChecked(TMT_MinimapAreaTooltip_DisplayNameInTitle)

        TMT_NamesUnderTitleCheckBox_GlobalName:SetChecked(TMT_Tooltips_DisplayNameUnderTitle)

        TMT_SetButtonState(TMT_DisplayNPCsNames, TMT_NpcsInTitleCheckBox_GlobalName)
        TMT_SetButtonState(TMT_DisplayNPCsNames, TMT_NPCsColorButton_GlobalName)

        TMT_SetButtonState(TMT_DisplaySpellsNames, TMT_SpellsInTitleCheckBox_GlobalName)
        TMT_SetButtonState(TMT_DisplaySpellsNames, TMT_SpellsClassColorCheckBox_GlobalName)
        TMT_SetButtonState(not TMT_SpellTooltip_UseClassColor and TMT_DisplaySpellsNames, TMT_SpellsColorButton_GlobalName)

        TMT_SetButtonState(TMT_DisplayItemsNames, TMT_ItemsInTitleCheckBox_GlobalName)
        TMT_SetButtonState(TMT_DisplayItemsNames, TMT_ItemsBlizzardColorCheckButton_GlobalName)
        TMT_SetButtonState(not TMT_ItemTooltip_UseBlizzardsColor and TMT_DisplayItemsNames, TMT_ItemsColorButton_GlobalName)
        
        TMT_SetButtonState(TMT_DisplayQuestsNames and not TMT_QuestsNames_UseBlizzardsColor, TMT_QuestsColorButton_GlobalName)
        TMT_SetButtonState(TMT_DisplayQuestsNames and not TMT_QuestsNames_UseBlizzardsColor and not QuestLogEx and not QuestGuru, TMT_highlightedEntryBGColor_GlobalName)
        TMT_SetButtonState(TMT_DisplayQuestsNames, TMT_QuestsBlizzardColorCheckButton_GlobalName)
        TMT_SetButtonState(TMT_DisplayQuestsNames and not QuestLogEx and not QuestGuru, TMT_QuestsTagsCheckButton_GlobalName)

        TMT_SetButtonState(TMT_DisplayAreasNames, TMT_AreasInTitleCheckBox_GlobalName)
    end
end

local function TMT_RestoreDefaults_QuestsFramePosition()
    TMT_QuestsFrameX = TMT_Defaults.questsFrameX
    TMT_QuestsFrameY = TMT_Defaults.questsFrameY
    TMT_ChangeFramePosition(TMT_QuestsFrame, "TOPLEFT", TMT_QuestsFrameParent, "TOPRIGHT", TMT_QuestsFrameX, TMT_QuestsFrameY)
    TMT_QuestsFrame_PositionChangedByUser = questsFrame_PositionChangedByUser
    TMT_QuestsFramePoint = TMT_Defaults.questsFramePoint
    TMT_QuestsFrameRelativePoint = TMT_Defaults.questsFrameRelativePoint
end

local function TMT_RestoreDefaults_Quests()
    TMT_RestoreDefaults_QuestsFramePosition()
    TMT_QuestColor = TMT_Defaults.questColor
    TMT_DisplayQuestsNames = TMT_Defaults.displayQuestsNames
    TMT_QuestsNamesToggle = TMT_Defaults.questsNamesToggle
    TMT_QuestsFrame_HighlightBackdropColor = TMT_Defaults.questsFrame_HighlightBackdropColor
    TMT_QuestsNames_UseBlizzardsColor = TMT_Defaults.questsNames_UseBlizzardsColor
    TMT_QuestsNames_DisplayQuestsTags = TMT_Defaults.questsNames_DisplayQuestsTags

    TMT_ConfigPanelSanityCheck()
end

local function TMT_RestoreDefaults_NPCs()
    TMT_NpcColor = TMT_Defaults.npcColor
    TMT_DisplayNPCsNames = TMT_Defaults.displayNPCsNames
    TMT_NPCTooltip_DisplayNameInTitle = TMT_Defaults.npcTooltip_displayNameInTitle

    TMT_ConfigPanelSanityCheck()
end

local function TMT_RestoreDefaults_Items()
    TMT_ItemColor = TMT_Defaults.itemColor
    TMT_ItemTooltip_DisplayNameInTitle = TMT_Defaults.itemTooltip_displayNameInTitle
    TMT_ItemTooltip_UseBlizzardsColor = TMT_Defaults.itemTooltip_UseBlizzardsColor
    TMT_DisplayItemsNames = TMT_Defaults.displayItemsNames

    TMT_ConfigPanelSanityCheck()
end

local function TMT_RestoreDefaults_Spells()
    TMT_SpellColor = TMT_Defaults.spellColor
    TMT_DisplaySpellsNames = TMT_Defaults.displaySpellsNames
    TMT_SpellTooltip_DisplayNameInTitle = TMT_Defaults.spellTooltip_displayNameInTitle
    TMT_SpellTooltip_UseClassColor = TMT_Defaults.spellTooltip_UseClassColor

    TMT_ConfigPanelSanityCheck()
end

local function TMT_RestoreDefaults_Areas()
    TMT_DisplayAreasNames = TMT_Defaults.displayAreasNames
    TMT_MinimapAreaTooltip_DisplayNameInTitle = TMT_Defaults.areaTooltip_DisplayNameInTitle

    TMT_ConfigPanelSanityCheck()
end

local function TMT_RestoreDefaults_Globals()
    TMT_Tooltips_DisplayNameUnderTitle = TMT_Defaults.tooltips_DisplayNameUnderTitle
end

-- Restores the addon default options
local function TMT_RestoreDefaults()
    TMT_RestoreDefaults_Globals()
    TMT_RestoreDefaults_Spells()
    TMT_RestoreDefaults_NPCs()
    TMT_RestoreDefaults_Items()
    TMT_RestoreDefaults_Quests()
    TMT_RestoreDefaults_Areas()

    TMT_ConfigPanelSanityCheck()
end

-- Create a checkbox
local function TMT_CheckboxTemplate(globalName, text, tooltip, hMargin, vMargin, parent)
    local checkBox = CreateFrame("CheckButton", globalName, parent, "ChatConfigCheckButtonTemplate")
    _G[checkBox:GetName() .. "Text"]:SetText(text); -- _G[] access the global namespace
    checkBox.tooltip = tooltip
    checkBox:SetPoint("TOPLEFT", hMargin, vMargin)

    return checkBox
end

-- Creates a default button
local function TMT_DefaultButton(parent, relativeAnchor, anchor, width, height, hMargin, vMargin, text, font, globalName)
    local button = CreateFrame("Button", globalName, parent, "UIPanelButtonTemplate")
    button:SetWidth(width)
    button:SetHeight(height)
    button:SetPoint(anchor, parent, relativeAnchor, hMargin, vMargin)

    local buttonText = button:CreateFontString(nil, "ARTWORK", font)
    buttonText:SetPoint("CENTER", button, "CENTER")
    buttonText:SetText(text)

    return button
end

local function TMT_QuestButton(width, height, hMargin, vMargin, text, font, parent, globalName)
    local button = TMT_DefaultButton(parent, "TOPLEFT", "TOPLEFT", width, height, hMargin, vMargin, text, font, globalName)
    return button
end

--[[
    This function is called when :
        - "ok" or "cancel" is clicked in the color picker frame
        - the color is changed in the color picker frame
        - the opacity is changed in the color pircker frame (not handled here)

    previousColor is set  when user clicks on "cancel", nil otherwise
]]
local function TMT_ColorCallBack(previousColor)
    local r, g, b, a

    if previousColor then
        r, g, b, a = unpack(previousColor)
    else
        r, g, b = ColorPickerFrame:GetColorRGB()
        a = 1 - OpacitySliderFrame:GetValue() -- slider values are reversed (0 at the top), weird
    end

    if TMT_ColorBeingChanged == "NPC" then 
        -- color picker works between 0 .. 1, I work between 0 .. 255
        TMT_NpcColor = {r * 255, g * 255, b * 255} 
    elseif TMT_ColorBeingChanged == "spell" then  
        TMT_SpellColor = {r * 255, g * 255, b * 255}
    elseif TMT_ColorBeingChanged == "item" then
        TMT_ItemColor = {r * 255, g * 255, b * 255}
    elseif TMT_ColorBeingChanged == "quest" then
        TMT_QuestColor = {r * 255, g * 255, b * 255}
        -- next 3 lines for conveniance while color picking
        TMT_QuestsFrameParent:Hide()
        TMT_QuestsFrameParent:Show()
        TMT_ChangeFramePosition(TMT_QuestsFrame_GlobalName, "TOPRIGHT", UIParent, "TOPRIGHT", 0, UIParent:GetHeight() / -4)
    elseif TMT_ColorBeingChanged == "highlightedEntryBG" then
        TMT_QuestsFrame_HighlightBackdropColor = {r * 255, g * 255, b * 255, a}
        -- next 3 lines for conveniance while color picking
        TMT_QuestsFrameParent:Hide()
        TMT_QuestsFrameParent:Show()
        TMT_ChangeFramePosition(TMT_QuestsFrame_GlobalName, "TOPRIGHT", UIParent, "TOPRIGHT", 0, UIParent:GetHeight() / -4)
    end

    -- Puts the frame displaying quests names back in place after color picking
    -- Reminder : this function is also called back when color picker is closed
    if not ColorPickerFrame:IsVisible() then
        TMT_QuestsFrameParent:Hide()

        if not TMT_QuestsFrame_PositionChangedByUser then
            TMT_ChangeFramePosition(TMT_QuestsFrame_GlobalName, TMT_QuestsFramePoint, TMT_QuestsFrameParent, TMT_QuestsFrameRelativePoint, TMT_QuestsFrameX, TMT_QuestsFrameY)
        else
            TMT_ChangeFramePosition(TMT_QuestsFrame_GlobalName, TMT_QuestsFramePoint, UIParent, TMT_QuestsFrameRelativePoint, TMT_QuestsFrameX, TMT_QuestsFrameY)
        end
    end
end

-- Called when color picker is shown
local function TMT_ShowColorPicker(rgba, colorBeingChanged, callBack)
    local r,g,b,a = unpack(rgba)
    -- I work between 0 and 255, color Picker works between 0 and 1
    r = r / 255
    g = g / 255
    b = b / 255

    TMT_ColorBeingChanged = colorBeingChanged

    ColorPickerFrame.func, ColorPickerFrame.opacityFunc, ColorPickerFrame.cancelFunc = 
        callBack, callBack, callBack
    ColorPickerFrame:SetColorRGB(r,g,b)
    ColorPickerFrame.hasOpacity = a ~= nil

    if ColorPickerFrame.hasOpacity then
        ColorPickerFrame.opacity = 1 - a -- slider values are reversed (0 at the top), weird
    end
    
    ColorPickerFrame.previousValues = {r,g,b,a}

    -- apparently needed to correctly throw "show" event
    ColorPickerFrame:Hide()
    ColorPickerFrame:Show()
end

local function TMT_GetClassColor()
    local localeClass, englishClass, classIndex = UnitClass("player")

    return TMT_Colors.classes[englishClass]
end

local function TMT_LanguageChange(dropdown, info)
    local previousLanguage = TMT_Language
    TMT_Language = info.arg1

    if previousLanguage ~= TMT_Language then
        --[[
            We have to reload to get all languages tables back because we delete 
            them to save memory if they are not the active language tables
        ]]
        ReloadUI()
        TMT_SetLanguage()
    end

    TMT_ConfigPanelSanityCheck()
end

-- Configuration panel
local function TMT_CreateConfigPanel()
    local configPanel = CreateFrame("FRAME", "TMT_ConfigPanel_GlobalName", UIParent)
    local lineWidth = 136
    local columnWidth = 180
    local firstColumn = TMT_HMargin
    local secondColumn = firstColumn + columnWidth
    local thirdColumn = secondColumn + columnWidth
    local firstLine = -TMT_VMargin
    local secondLine = firstLine - lineWidth
    local thirdLine = secondLine - lineWidth
    local fourthLine = thirdLine - lineWidth
    local checkBoxSpace = 24
    local buttonSpace = 16
    local bigButtonSpace = 32
    local titleSpace = 32
    local textLineSpace = 16
    local tab = 15
    local HPos
    local VPos
    configPanel.name = "The Mother Tongue" -- The name in the "Add-ons" tab of interface
    InterfaceOptions_AddCategory(configPanel)

    -- Title
    HPos = firstColumn
    VPos = firstLine
    local configPanelTitle = configPanel:CreateFontString(nil, "ARTWORK", "GameFontNormalLarge")
    configPanelTitle:SetPoint("TOPLEFT", HPos, VPos)
    configPanelTitle:SetText("The Mother Tongue")

    -- Version
    HPos = firstColumn
    VPos = firstLine - textLineSpace
    local configPanelSubtitle = configPanel:CreateFontString(nil, "ARTWORK", "GameFontNormalSmall")
    configPanelSubtitle:SetPoint("TOPLEFT", HPos, VPos)
    configPanelSubtitle:SetText(TMT.version)

    if TMT_NoLanguagePack then
        HPos = 0
        VPos = secondLine
        local globalCat = configPanel:CreateFontString(nil, "ARTWORK", "GameFontNormalLarge")
        globalCat:SetPoint("TOP", HPos, VPos)
        globalCat:SetText("You have no TMT language pack loaded !")

        HPos = 0
        VPos = secondLine - titleSpace
        local globalCat = configPanel:CreateFontString(nil, "ARTWORK", "GameFontNormalLarge")
        globalCat:SetPoint("TOP", HPos, VPos)
        globalCat:SetText("You can get them like any other AddOn.")


        HPos = 0
        VPos = secondLine - titleSpace - textLineSpace - textLineSpace - textLineSpace
        local globalCat = configPanel:CreateFontString(nil, "ARTWORK", "GameFontNormalLarge")
        globalCat:SetPoint("TOP", HPos, VPos)
        globalCat:SetText("Available packs : ")

        HPos = 0
        VPos = secondLine - titleSpace - textLineSpace - textLineSpace - textLineSpace - textLineSpace - textLineSpace
        local globalCat = configPanel:CreateFontString(nil, "ARTWORK", "GameFontNormalLarge")
        globalCat:SetPoint("TOP", HPos, VPos)
        globalCat:SetText("TheMotherTongue_LanguagePack_en (English)")

        HPos = 0
        VPos = secondLine - titleSpace - textLineSpace - textLineSpace - textLineSpace - textLineSpace - textLineSpace - textLineSpace - textLineSpace
        local globalCat = configPanel:CreateFontString(nil, "ARTWORK", "GameFontNormalLarge")
        globalCat:SetPoint("TOP", HPos, VPos)
        globalCat:SetText("TheMotherTongue_LanguagePack_de (German)")
    else
        --language selection title
        HPos = firstColumn - tab - tab
        VPos = firstLine
        local languageSelectionTitle = configPanel:CreateFontString(nil, "ARTWORK", "GameFontNormal")
        languageSelectionTitle:SetPoint("TOPRIGHT", HPos, VPos)
        languageSelectionTitle:SetText("Language")

        -- Language selection dropdown
        HPos = firstColumn - tab
        VPos = firstLine - textLineSpace
        local settings = {}
        settings.parent = configPanel
        settings.name = "TMT_LanguageSelectionDropdown_GlobalName"
        settings.values = TMT_AvailableLanguages
        settings.changeHandler = TMT_LanguageChange

        local dropdown = TMT_LanguageDropdown(settings)
        dropdown:SetPoint("TOPRIGHT", HPos, VPos)

        -- Restore Settings button
        HPos = firstColumn + tab
        VPos = firstLine - titleSpace
        local defaultButton = TMT_DefaultButton(configPanel, "TOPLEFT", "TOPLEFT", 128, 24, HPos, VPos, "Reset Settings", "GameFontNormal")
        defaultButton:SetScript("OnClick", function(self) TMT_RestoreDefaults() end)

        -- Global parameters title
        HPos = firstColumn
        VPos = firstLine - titleSpace - bigButtonSpace - textLineSpace
        local globalCat = configPanel:CreateFontString(nil, "ARTWORK", "GameFontNormal")
        globalCat:SetPoint("TOPLEFT", HPos, VPos)
        globalCat:SetText("Global settings")

        -- Globals parameters
        HPos = firstColumn + tab
        VPos = firstLine - titleSpace - bigButtonSpace - textLineSpace - textLineSpace
        local namesUnderTitleCheckBox = TMT_CheckboxTemplate("TMT_NamesUnderTitleCheckBox_GlobalName", " Display names under title",  "If checked, names set to be displayed in the title are displayed under it instead", HPos, VPos, configPanel)
        namesUnderTitleCheckBox:SetScript("OnClick", function(self) TMT_Tooltips_DisplayNameUnderTitle = self:GetChecked(); TMT_ConfigPanelSanityCheck() end)

        -- NPCs subcategory title
        HPos = firstColumn
        VPos = secondLine
        local npcCat = configPanel:CreateFontString(nil, "ARTWORK", "GameFontNormal")
        npcCat:SetPoint("TOPLEFT", HPos, VPos)
        npcCat:SetText("NPCs settings")

        -- Reset settings button
        HPos = firstColumn + tab * 6
        VPos = secondLine
        local npcsDefaultButton = TMT_DefaultButton(configPanel, "TOPLEFT", "TOPLEFT", 48, 16, HPos, VPos, "Reset", "GameFontNormalSmall")
        npcsDefaultButton:SetScript("OnClick", function(self) TMT_RestoreDefaults_NPCs() end)

        -- NPCs names checkbox
        HPos = firstColumn + tab
        VPos = secondLine - textLineSpace
        local npcsNamesCheckBox = TMT_CheckboxTemplate("TMT_NpcsNamesCheckBox_GlobalName", " NPCs names",  "If checked, displays the English name of the NPC the mouse is hovering above in the tooltip", HPos, VPos, configPanel)
        npcsNamesCheckBox:SetScript("OnClick", function(self) TMT_DisplayNPCsNames = self:GetChecked(); TMT_ConfigPanelSanityCheck() end)
        
        -- NPCs in title checkbox
        HPos = firstColumn + tab
        VPos = secondLine - textLineSpace - checkBoxSpace
        local npcsInTitleCheckBox = TMT_CheckboxTemplate("TMT_NpcsInTitleCheckBox_GlobalName", " In title", "If checked, displays the name in the tooltip title. If not, displays it on a new line at the end of the tooltip.", HPos, VPos, configPanel)
        npcsInTitleCheckBox:SetScript("OnClick", function(self) TMT_NPCTooltip_DisplayNameInTitle = self:GetChecked(); TMT_ConfigPanelSanityCheck() end)

        -- NPCs color button
        HPos = firstColumn + tab
        VPos = secondLine - textLineSpace - checkBoxSpace - checkBoxSpace
        local npcsColorButton = TMT_DefaultButton(configPanel, "TOPLEFT", "TOPLEFT", 80, 19, HPos, VPos, "Name Color", "GameFontNormalSmall", "TMT_NPCsColorButton_GlobalName")
        npcsColorButton:SetScript("OnClick", function(self) TMT_ShowColorPicker(TMT_NpcColor, "NPC", TMT_ColorCallBack) end)

        -- Spells subcategory title
        HPos = secondColumn
        VPos = secondLine
        local spellsCat = configPanel:CreateFontString(nil, "ARTWORK", "GameFontNormal")
        spellsCat:SetPoint("TOPLEFT", HPos, VPos)
        spellsCat:SetText("Spells settings")

        -- Reset settings button
        HPos = secondColumn + tab * 6
        VPos = secondLine
        local spellsDefaultButton = TMT_DefaultButton(configPanel, "TOPLEFT", "TOPLEFT", 48, 16, HPos, VPos, "Reset", "GameFontNormalSmall")
        spellsDefaultButton:SetScript("OnClick", function(self) TMT_RestoreDefaults_Spells() end)

        -- Spells names checkbox
        HPos = secondColumn + tab
        VPos = secondLine - textLineSpace
        local spellsNamesCheckBox = TMT_CheckboxTemplate("TMT_SpellsNamesCheckBox_GlobalName", " Spells names", "If checked, displays the English name of the spell the mouse is hovering above in the tooltip", HPos, VPos, configPanel)
        spellsNamesCheckBox:SetScript("OnClick", function(self) TMT_DisplaySpellsNames = self:GetChecked(); TMT_ConfigPanelSanityCheck() end)

        -- Spells in title checkbox
        HPos = secondColumn + tab
        VPos = secondLine - textLineSpace - checkBoxSpace
        local spellsInTitleCheckBox = TMT_CheckboxTemplate("TMT_SpellsInTitleCheckBox_GlobalName", " In title", "If checked, displays the name in the tooltip title. If not, displays it on a new line at the end of the tooltip.", HPos, VPos, configPanel)
        spellsInTitleCheckBox:SetScript("OnClick", function(self) TMT_SpellTooltip_DisplayNameInTitle = self:GetChecked(); TMT_ConfigPanelSanityCheck() end)

        -- Spells class color checkBox
        HPos = secondColumn + tab
        VPos = secondLine - textLineSpace - checkBoxSpace - checkBoxSpace
        local spellsClassColorCheckBox = TMT_CheckboxTemplate("TMT_SpellsClassColorCheckBox_GlobalName", " Use class color", "If checked, uses your class color", HPos, VPos, configPanel)
        spellsClassColorCheckBox:SetScript("OnClick", function(self) TMT_SpellTooltip_UseClassColor = self:GetChecked(); TMT_ConfigPanelSanityCheck() end)

        -- Spells color button
        HPos = secondColumn + tab
        VPos = secondLine - textLineSpace - checkBoxSpace - checkBoxSpace - checkBoxSpace
        local spellsColorButton = TMT_DefaultButton(configPanel, "TOPLEFT", "TOPLEFT", 80, 19, HPos, VPos, "Name Color", "GameFontNormalSmall", "TMT_SpellsColorButton_GlobalName")
        spellsColorButton:SetScript("OnClick", function(self) TMT_ShowColorPicker(TMT_SpellColor, "spell", TMT_ColorCallBack) end)

        -- Items subcategory title
        HPos = thirdColumn
        VPos = secondLine
        local itemsCat = configPanel:CreateFontString(nil, "ARTWORK", "GameFontNormal")
        itemsCat:SetPoint("TOPLEFT", HPos, VPos)
        itemsCat:SetText("Items settings")

        -- Reset settings button
        HPos = thirdColumn + tab * 6
        VPos = secondLine
        local itemsDefaultButton = TMT_DefaultButton(configPanel, "TOPLEFT", "TOPLEFT", 48, 16, HPos, VPos, "Reset", "GameFontNormalSmall")
        itemsDefaultButton:SetScript("OnClick", function(self) TMT_RestoreDefaults_Items() end)

        -- Items names checkbox
        HPos = thirdColumn + tab
        VPos = secondLine - textLineSpace
        local itemsNamesCheckBox = TMT_CheckboxTemplate("TMT_ItemsNamesCheckBox_GlobalName", " Items names", "If checked, displays the English name of the item the mouse is hovering above in the tooltip", HPos, VPos, configPanel)
        itemsNamesCheckBox:SetScript("OnClick", function(self) TMT_DisplayItemsNames = self:GetChecked(); TMT_ConfigPanelSanityCheck() end)

        -- Items in title checkbox
        HPos = thirdColumn + tab
        VPos = secondLine - textLineSpace - checkBoxSpace
        local itemsInTitleCheckBox = TMT_CheckboxTemplate("TMT_ItemsInTitleCheckBox_GlobalName", " In title", "If checked, displays the name in the tooltip title. If not, displays it on a new line at the end of the tooltip.", HPos, VPos, configPanel)
        itemsInTitleCheckBox:SetScript("OnClick", function(self) TMT_ItemTooltip_DisplayNameInTitle = self:GetChecked(); TMT_ConfigPanelSanityCheck() end)

        -- Items tooltip blizzard's color checkbox
        HPos = thirdColumn + tab
        VPos = secondLine - textLineSpace - checkBoxSpace - checkBoxSpace
        local itemsBlizzardColorCheckButton = TMT_CheckboxTemplate("TMT_ItemsBlizzardColorCheckButton_GlobalName", " Blizzard's colors", "If checked, uses Blizzard's colors to display items English names", HPos, VPos, configPanel)
        itemsBlizzardColorCheckButton:SetScript("OnClick", function(self) TMT_ItemTooltip_UseBlizzardsColor = self:GetChecked(); TMT_ConfigPanelSanityCheck() end)

        -- Items color button
        HPos = thirdColumn + tab
        VPos = secondLine - textLineSpace - checkBoxSpace - checkBoxSpace - checkBoxSpace
        local itemsColorButton = TMT_DefaultButton(configPanel, "TOPLEFT", "TOPLEFT", 80, 19, HPos, VPos, "Name Color", "GameFontNormalSmall", "TMT_ItemsColorButton_GlobalName")
        itemsColorButton:SetScript("OnClick", function(self) TMT_ShowColorPicker(TMT_ItemColor, "item", TMT_ColorCallBack) end)

        -- Quests subcategory title
        HPos = firstColumn
        VPos = thirdLine
        local questsCat = configPanel:CreateFontString(nil, "ARTWORK", "GameFontNormal")
        questsCat:SetPoint("TOPLEFT", HPos, VPos)
        questsCat:SetText("Quests settings")

        -- Reset settings button
        HPos = firstColumn + tab * 6.5
        VPos = thirdLine
        local questsDefaultButton = TMT_DefaultButton(configPanel, "TOPLEFT", "TOPLEFT", 48, 16, HPos, VPos, "Reset", "GameFontNormalSmall")
        questsDefaultButton:SetScript("OnClick", function(self) TMT_RestoreDefaults_Quests() end)

        -- Quests checkbox
        HPos = firstColumn + tab
        VPos = thirdLine - textLineSpace
        local questCheckButton = TMT_CheckboxTemplate("TMT_QuestsNamesCheckButton_GlobalName", " Quests names", "If checked, displays the English name of the quests currently in your questlog", HPos, VPos, configPanel)
        questCheckButton:SetScript("OnClick", function(self) TMT_DisplayQuestsNames = self:GetChecked(); TMT_ConfigPanelSanityCheck() end)

        -- Quests tags checkbox
        HPos = firstColumn + tab
        VPos = thirdLine - textLineSpace - checkBoxSpace
        local questsTagsCheckButton = TMT_CheckboxTemplate("TMT_QuestsTagsCheckButton_GlobalName", " Quests tags", "If checked, displays the quests tags (Dungeon, etc.)", HPos, VPos, configPanel)
        questsTagsCheckButton:SetScript("OnClick", function(self) TMT_QuestsNames_DisplayQuestsTags = self:GetChecked(); TMT_ConfigPanelSanityCheck() end)

        -- Quests Blizz's colors checkbox
        HPos = firstColumn + tab
        VPos = thirdLine - textLineSpace - checkBoxSpace - checkBoxSpace
        local questsBlizzardColorCheckButton = TMT_CheckboxTemplate("TMT_QuestsBlizzardColorCheckButton_GlobalName", " Blizzard's colors", "If checked, uses Blizzard's colors to display quests English names", HPos, VPos, configPanel)
        questsBlizzardColorCheckButton:SetScript("OnClick", function(self) TMT_QuestsNames_UseBlizzardsColor = self:GetChecked(); TMT_ConfigPanelSanityCheck() end)

        -- Quests color button
        HPos = firstColumn + tab
        VPos = thirdLine - textLineSpace - checkBoxSpace - checkBoxSpace - checkBoxSpace
        local questsColorButton = TMT_DefaultButton(configPanel, "TOPLEFT", "TOPLEFT", 80, 19, HPos, VPos, "Name Color", "GameFontNormalSmall", "TMT_QuestsColorButton_GlobalName")
        questsColorButton:SetScript("OnClick", function(self) TMT_ShowColorPicker(TMT_QuestColor, "quest", TMT_ColorCallBack) end)

        -- Highlighted entry background color
        HPos = firstColumn + tab
        VPos = thirdLine - textLineSpace - checkBoxSpace - checkBoxSpace - checkBoxSpace - checkBoxSpace
        local highlightedEntryBGColor = TMT_DefaultButton(configPanel, "TOPLEFT", "TOPLEFT", 95, 19, HPos, VPos, "Highlight color", "GameFontNormalSmall", "TMT_highlightedEntryBGColor_GlobalName")
        highlightedEntryBGColor:SetScript("OnClick", function(self) TMT_ShowColorPicker(TMT_QuestsFrame_HighlightBackdropColor, "highlightedEntryBG", TMT_ColorCallBack) end)

        -- Areas subcategory title
        HPos = secondColumn
        VPos = thirdLine
        local areasCat = configPanel:CreateFontString(nil, "ARTWORK", "GameFontNormal")
        areasCat:SetPoint("TOPLEFT", HPos, VPos)
        areasCat:SetText("Areas settings")

        -- Reset settings button
        HPos = secondColumn + tab * 6
        VPos = thirdLine
        local areasDefaultButton = TMT_DefaultButton(configPanel, "TOPLEFT", "TOPLEFT", 48, 16, HPos, VPos, "Reset", "GameFontNormalSmall")
        areasDefaultButton:SetScript("OnClick", function(self) TMT_RestoreDefaults_Areas() end)

        -- Areas names checkbox
        HPos = secondColumn + tab
        VPos = thirdLine - textLineSpace
        local areasNamesCheckButton = TMT_CheckboxTemplate("TMT_AreasNamesCheckBox_GlobalName", " Areas names", "If checked, displays the English name of the area in the tooltip when the mouse is hovering above the minimap area name", HPos, VPos, configPanel)
        areasNamesCheckButton:SetScript("OnClick", function(self) TMT_DisplayAreasNames = self:GetChecked(); TMT_ConfigPanelSanityCheck() end)

        -- Areas in title checkbox
        HPos = secondColumn + tab
        VPos = thirdLine - textLineSpace - checkBoxSpace
        local areasInTitleCheckBox = TMT_CheckboxTemplate("TMT_AreasInTitleCheckBox_GlobalName", " In title", "If checked, displays the name in the tooltip title. If not, displays it on a new line at the end of the tooltip.", HPos, VPos, configPanel)
        areasInTitleCheckBox:SetScript("OnClick", function(self) TMT_MinimapAreaTooltip_DisplayNameInTitle = self:GetChecked(); TMT_ConfigPanelSanityCheck() end)

        -- Notes
        HPos = firstColumn
        VPos = fourthLine - checkBoxSpace
        local notesTitle = configPanel:CreateFontString(nil, "ARTWORK", "GameFontNormal")
        notesTitle:SetPoint("TOPLEFT", HPos, VPos)
        notesTitle:SetText("Notes")

        HPos = firstColumn + tab
        VPos = fourthLine - checkBoxSpace - textLineSpace
        local notes1 = configPanel:CreateFontString(nil, "ARTWORK", "GameFontNormalSmall")
        notes1:SetPoint("TOPLEFT", HPos, VPos)
        notes1:SetText("- alt + click on a quest name writes it in chat")

        HPos = firstColumn + tab
        VPos = fourthLine - checkBoxSpace - textLineSpace - textLineSpace
        local notes1 = configPanel:CreateFontString(nil, "ARTWORK", "GameFontNormalSmall")
        notes1:SetPoint("TOPLEFT", HPos, VPos)
        notes1:SetText("- You can get a TMT language pack like any other AddOn. Available packs : ")

        HPos = firstColumn + tab + tab
        VPos = fourthLine - checkBoxSpace - textLineSpace - textLineSpace - textLineSpace
        local notes2 = configPanel:CreateFontString(nil, "ARTWORK", "GameFontNormalSmall")
        notes2:SetPoint("TOPLEFT", HPos, VPos)
        notes2:SetText("- TheMotherTongue_LanguagePack_en (English)")

        HPos = firstColumn + tab + tab
        VPos = fourthLine - checkBoxSpace - textLineSpace - textLineSpace - textLineSpace - textLineSpace
        local notes3 = configPanel:CreateFontString(nil, "ARTWORK", "GameFontNormalSmall")
        notes3:SetPoint("TOPLEFT", HPos, VPos)
        notes3:SetText("- TheMotherTongue_LanguagePack_de (German)")
    end

    TMT_ConfigPanelSanityCheck()
end

-- Slash commands code
local function TMT_Commands(msg, editbox)
    if msg == "sc" or msg == "sanitycheck" then
        print("Sanity checked")
    elseif msg == "npcsnameson" then
        TMT_DisplayNPCsNames = true;
    elseif msg == "npcsnamesoff" then
        TMT_DisplayNPCsNames = false;
    elseif msg == "npcnametop" then
        TMT_NPCTooltip_DisplayNameInTitle = true
    elseif msg == "npcnamebottom" then
        TMT_NPCTooltip_DisplayNameInTitle = false
    elseif msg == "questsnameson" then
        TMT_DisplayQuestsNames = true;
    elseif msg == "questsnamesoff" then
        TMT_DisplayQuestsNames = false;
    elseif msg == "spellsnameson" then
        TMT_DisplaySpellsNames = true;
    elseif msg == "spellsnamesoff" then
        TMT_DisplaySpellsNames = false;
    elseif msg == "spellnametop" then
        TMT_SpellTooltip_DisplayNameInTitle = true
    elseif msg == "spellnamebottom" then
        TMT_SpellTooltip_DisplayNameInTitle = false
    elseif msg == "version" or msg == "v" then
        print(TMT_Colors.orange .. "TMT " .. TMT_Colors.green .. TMT.version)
    --elseif msg == "getspellsnames" then
        --TMT_GetSpellsNames()
    elseif msg == "gettalents" then
        TMT_GetTalents()
    elseif msg == "getareas" then
        TMT_GetAreas()
    elseif msg == "getzones" then
        TMT_GetZones()
    elseif msg == "getquests" then
        TMT_GetQuests()
    else 
        -- called twice to work around a BLizzard's bug
        InterfaceOptionsFrame_OpenToCategory(TMT_ConfigPanel_GlobalName)
        InterfaceOptionsFrame_OpenToCategory(TMT_ConfigPanel_GlobalName)
    end

    TMT_ConfigPanelSanityCheck()
end

SLASH_TMT1 = "/tmt"
SlashCmdList["TMT"] = TMT_Commands

local function TMT_DetectLanguagesPacks()
    local packageFound = false
    local firstPackageFound

    for key,val in pairs(TMT_Languages) do
        local language = _G["TheMotherTongue_LanguagePack_" .. key];
        if language then
            if(not packageFound) then
                firstPackageFound = key
            end
            packageFound = true
            TMT_AvailableLanguages[key] = val
        end
    end
    
    -- Sets the active language
    if packageFound then
        if not TMT_AvailableLanguages[TMT_Language] then
            -- We had a language pack before, but it's no longer there
            if TMT_AvailableLanguages.en then
                TMT_Language = "en"
            else
                TMT_Language = firstPackageFound
            end
        end

        TMT_SetLanguage()
    else
        TMT_Language = ""
        print(TMT_Colors.orange .. "TMT: " .. TMT_Colors.red 
            .. "No TMT language pack detected ! Type '/tmt' to learn how to get them")
    end

    -- Deactivate / activate the displays
    TMT_DisplayNPCsNames = packageFound
    TMT_DisplayQuestsNames = packageFound
    TMT_DisplaySpellsNames = packageFound
    TMT_DisplayItemsNames = packageFound
    TMT_DisplayAreasNames = packageFound
    TMT_NoLanguagePack = not packageFound
end

--[[ 
    - Algorithm used to find the English name corresponding to an ID.
    - idNamePairs is the list of {id, "name"} pairs where to search for the id.
    - TO DO : order the lists to use a better search algorithm.
]]
local function TMT_FindNameFromId(idToFind, idNamePairs)
    local name

    for _, pair in ipairs(idNamePairs) do
        if pair[1] == idToFind then
            name = pair[2]
            break
        end
    end

    return name
end

local function TMT_DisplayUpdateMessage()
    local updateMessage = TMT_Colors.orange .. "TMT: Language packs are now supported. Don't forget to install one ! Type '/tmt' to learn how to get them." 

    if TMT.version ~= TMT_LastUpdateMessageVersion then
        print(updateMessage)
        TMT_LastUpdateMessageVersion = TMT.version
    end
end


local function TMT_TooltipDisplay(tooltip, name, inTitle, color)
    if name then
        if not color then
            color = TMT_Colors.orange
        end

        -- Doesn't display if the name in English is the same
        if GameTooltipTextLeft1:GetText() ~= name then
            if inTitle then -- User option
                local placement = "|r" .. TMT_Colors.lightGrey .. " - "
                
                if TMT_Tooltips_DisplayNameUnderTitle then
                    placement = "|n"
                end

                tooltip:AppendText(placement .. color .. name)
            else 
                tooltip:AddLine(color .. name)
            end
        end
    end
end

--[[
    Parse the GUID to get the actual ID of the unit.
    A unit GUID is of the form : Creature-0-4480-1-244-2955-0006FF9B5.
    The ID we want here would be 2955 (a Plainstrider btw).
]]
local function TMT_GetUnitId(guid)
    if guid:match("^%a+") ~= "Player" then
        return tonumber(guid:match("-(%d+)-%x+$"), 10)
    end
end

local function TMT_FindNpcOriginalName(unit)
    local unitName
    local color = TMT_RgbToHex(TMT_NpcColor)

    if unit then -- needed in SuF and LunaUF to not trigger a LUA bug on ToT and ToToT
        local guid = UnitGUID(unit)
        local idToFind = TMT_GetUnitId(guid)
        unitName = TMT_FindNameFromId(idToFind, TMT.npcs)
    end

    return unitName, color
end

local function TMT_TooltipTest()
    print(TMT_Colors.orange .. "TMT_Tooltiptest")
    GameTooltip:SetText("This is a sample tooltip with wrapping text.", 0.5, 0.5, 0.5,  0.75, true);
    GameTooltip:Show();
end

--[[
    A few words on the pool. Before version 1.1.0, I displayed only a text in
    the questsFrame. So I would reset the text everytime and set it to it's new value
    when the entries in the quests log changed (picking a new quest, collapsing/extending groups etc).
    Since 1.1.0, I use buttons instead of just the text for quests names, so I couldn't only use a plain text.
    The thing is, you can't delete a UI element such as a button when they are created.
    So I had to make a pool of questLog entries. When a new questLog entry appears, meaning
    the player picked up a new quest, it's added to the pool. When a quests disapears from
    the questsLog, it's still in the pool (can't delete). So I have to check
    the questlog pool (current entries in the questlog) and only show these ones and hide the rest
    of the questsFrame pool (all entries that appearead in the questlog since the log in).
]]
local function TMT_QuestsFramePoolDisplay(buttonWidth)
    -- First hides all the elements currently existing
    for index, element in pairs(TMT_QuestsFramePool) do
        element:ClearAllPoints()
        element:Hide()
    end

    -- Then shows only those currently in the quests log
    for index,idInPool in ipairs(TMT_QuestsLogEntriesPool) do
        if TMT_QuestsFramePool[idInPool] then
            local hMargin = TMT_HMargin

            if string.match(idInPool, "quest:") then
                hMargin = hMargin / 2

                -- Sets the position of the quest tag
                local questID = string.match(TMT_QuestsFramePool[idInPool]:GetName(), "quest: (%d+)_")

                    local tag = _G[questID .. "_Tag_GlobalName"]
                    if tag then
                        if TMT_QuestsNames_DisplayQuestsTags and not QuestGuru and not QuestLogEx then
                            tag:SetPoint("TOPRIGHT", TMT_QuestsFramePool[idInPool], "TOPRIGHT", -10, 0)
                        else
                            tag:ClearAllPoints()
                        end
                    end 
            end
            
            TMT_QuestsFramePool[idInPool]:SetWidth(buttonWidth)
            TMT_QuestsFramePool[idInPool]:SetPoint("TOPLEFT", hMargin, TMT_VMargin * index * -0.8 - TMT_VMargin)
            TMT_QuestsFramePool[idInPool]:Show()
        end
    end
end

-- Just switch between highlight and normal font on mouse hover
local function TMT_MousePassingByButton(self, color)
    -- if it's the selected entry, doesn't mess with the text color to keep the highlight

    if self ~= TMT_HighlightedEntry then
        TMT_SetButtonTextColor(self, color)
    end
end

local function TMT_SelectQuestLogEntry(ourIndex, isQuest)
    --[[
        Quest log entries have global names of the form : "QuestLogTitle[index]".
        The problem is that QuestLogFrame only has 6 entries children (only 6 displayed at the same time), 
        and the quests and headers are dynamically (I think) put in those children when the user scrolls
        through the entries. So, to select an entry that has an index > 6, we have to scroll accordingly. Same
        if we want to select an entry that is above the current 6 entries displayed (meaning the list is scrolled down).
    ]]
    if QuestLogEx or QuestGuru then
        --[[
            Selecting the entry is enough to show the quest, but not enough to
            expand/collapse headers or trigger shift + click. The thing is, since
            those addons don't switch the entries put in the 6 children
            (because they never scroll, since they show all quests at once), I can't
            trigger the click further than the first 6 entries. So I just don't trigger it.
            So those addons just get text for headers, not buttons.
        ]]
        SelectQuestLogEntry(ourIndex)
        TMT_QuestsFrameParent:Hide()
        TMT_QuestsFrameParent:Show()
    else
        local logIndex = ourIndex
        local scrollIncrement = 16 -- scroll goes up/down by increments of 16, not one
        local numberOfScrolls = 0

        -- we don't want to mess with the scroll if there's no scroll available to the user
        if GetNumQuestLogEntries() > 6 then
            if isQuest then 
                local upperVisibleBound = 6 + math.floor((QuestLogListScrollFrame:GetVerticalScroll() / scrollIncrement) + 0.5)
                local lowerVisibleBound = upperVisibleBound - 5 -- minus 5 because index starts at 1
            
                --[[
                    The behaviour when a scroll happens :
                    - If we need to scroll up, quest is shown at the top
                    - If we need to scroll down, quest is shown at the bottom
                ]]
                if ourIndex > upperVisibleBound or ourIndex < lowerVisibleBound then -- if it's not actually already visible
                    if ourIndex < lowerVisibleBound then -- we go up
                        numberOfScrolls = ourIndex - 1
                        logIndex = 1 -- we scrolled exactly to put the needed quest at the top
                    elseif ourIndex > upperVisibleBound then -- we go down
                        numberOfScrolls = ourIndex - 6
                        logIndex = 6 -- we scrolled exactly to put the needed quest at the bottom
                    end

                    QuestLogListScrollFrame:SetVerticalScroll(numberOfScrolls  * scrollIncrement)
                else -- didn't scroll
                    logIndex = ourIndex - (math.floor((QuestLogListScrollFrame:GetVerticalScroll() / scrollIncrement) + 0.5))
                end
            else
                --[[
                    The behaviour when a click on a header happens :
                    - Expands it and shows it at the top
                ]]

                numberOfScrolls = logIndex - 1
                logIndex = 1 -- we scrolled exactly to put the needed header at the top, we select the first quest
                QuestLogListScrollFrame:SetVerticalScroll(numberOfScrolls  * scrollIncrement)    
            end
        end

        if isQuest then -- We don't want to collapse the headers 
            SelectQuestLogEntry(logIndex) -- selects the entry but does nothing else
            _G["QuestLogTitle" .. logIndex]:Click()
        else -- just in case the header was collapsed
            ExpandQuestHeader(ourIndex)
            SelectQuestLogEntry(logIndex + 1) -- selects the first quest under the header
            _G["QuestLogTitle" .. logIndex + 1]:Click()
        end
    end
end

local function TMT_QuestsFrameButtonClick(self, isQuest)
    local buttonId = self:GetName()

    if IsAltKeyDown() then -- just print the English name in chat
        -- retain only the name of the quest
        if isQuest then
            -- id of the form quest:questId_name_level_GlobalName
            local name = string.match(buttonId, "%d+_(.+)_%d+")
            TMT_WriteInChat(name)
        end
    else -- Imulate the click on the scorresponding entry in the questLog
        local indexInLog

        -- we retrieve the index, in the questLog, of the quest corresponding to the button
        for index, logId in ipairs(TMT_QuestsLogEntriesPool) do
            if buttonId == (logId .. "_GlobalName") then
                indexInLog = index
                break
            end
        end

        TMT_SelectQuestLogEntry(indexInLog, isQuest)
    end
end

local function TMT_HighlightEntry(entryID)
    if not QuestLogEx and not QuestGuru then
        local title, level, suggestedGroup, isHeader, isCollapsed, isComplete, frequency, questID, startEvent, displayQuestID, isOnMap, hasLocalPOI, isTask, isStory = GetQuestLogTitle(entryID)
        local r, g, b, a = unpack(TMT_QuestsFrame_HighlightBackdropColor)

        if TMT_QuestsNames_UseBlizzardsColor then
            local difficultyColor = GetQuestDifficultyColor(level)
            r = math.floor((difficultyColor.r * 255) + 0.5)
            g = math.floor((difficultyColor.g * 255) + 0.5)
            b = math.floor((difficultyColor.b * 255) + 0.5)
            a = 0.7
        end

        if not isHeader then
            local entryInPoolID = "quest: ".. questID .. "_" .. TMT_FindNameFromId(questID, TMT.quests) .. "_" .. level
            local entry = TMT_QuestsFramePool[entryInPoolID]
            local entryName = ""
            
            for i, button in pairs(TMT_QuestsFramePool) do
                if string.find(button:GetName(), "quest") then
                    --[[
                        We put every button in the normal state to unselect the 
                        previous selected entry.
                        We do this on every button and not only on the previous 
                        selected one just in case a highlighted button was put out
                        of the current displayed pool while it was the selected button.
                    ]]
                    
                    local questLevel = string.match(button:GetName(), "_(%d+)")
                    TMT_SetButtonTextColor(button, TMT_GetQuestColor(questLevel))

                    local name = string.sub(button:GetText(), 11) -- removes the previous color

                    -- remove the backdrop
                    button:SetBackdrop(nil)
                end
            end

            if entry then -- entry is nil on reload and quests names are toggled off
                entry:SetBackdrop( { 
                    bgFile = "Interface/QUESTFRAME/UI-QuestLogTitleHighlight",
                    edgeFile = nil, --"Interface/DialogFrame/UI-DialogBox-Border",
                    tile = false, 
                    tileSize = 0,
                    edgeSize = 1,
                    insets = {left = 0, right = 8, top = 0, bottom = 0 }
                });
                entry:SetBackdropColor(r / 255, g / 255, b / 255, a)
                TMT_SetButtonTextColor(entry, TMT_Colors.white)
                TMT_HighlightedEntry = entry
            end
        end
    end
end

local function TMT_SetQuestsTags(previousMaxWidth)
    local longestTagWidth = 0

    for i, inPoolId in ipairs(TMT_QuestsLogEntriesPool) do
        local button = TMT_QuestsFramePool[inPoolId]
        
        if string.find(button:GetName(), "quest") then -- No tag on headers
            local title, level, suggestedGroup, isHeader, isCollapsed, isComplete, frequency, questID, startEvent, displayQuestID, isOnMap, hasLocalPOI, isTask, isStory = GetQuestLogTitle(i)
            local color = string.sub(button:GetText(), 0, 10)
            local questID = string.match(button:GetName(), "quest: (%d+)_")
            local tagID = GetQuestTagInfo(questID)
            local tagName = ""

            -- Dungeon, elite, etc
            if tagID then
                tagName = "("

                -- translates the tag in english
                for _, ourTag in ipairs(TMT_QuestsTags) do
                    if ourTag[1] == tagID then
                        tagName = tagName .. ourTag[2]
                        break
                    end
                end
                tagName = tagName .. ")"
            end

            -- quest completion
            if isComplete then 
                if isComplete == 1 then
                    tagName = "(Completed)" -- if completed, removes all other tags
                elseif isComplete == -1 then
                    tagName = "(Failed)" -- if failed, removes all other tags
                end
            end

            local tagFontString = _G[questID .. "_Tag_GlobalName"]

            if not tagFontString then -- if the tag wasn't already set for this quest
                tagFontString = button:CreateFontString(questID .. "_Tag_GlobalName", "ARTWORK", "QuestFontNormalSmall")
            end

            tagFontString:SetText(color .. tagName)

            if TMT_GetStringWidth(tagName) > longestTagWidth then
                longestTagWidth = TMT_GetStringWidth(tagName)
            end
        end
    end

    return previousMaxWidth + longestTagWidth
end

local function TMT_FindQuestsOriginalNames()
    local questsLogEntries = GetNumQuestLogEntries()
    
    if questsLogEntries > 0 then
        TMT_QuestsToggle_GlobalName:Show()
        TMT_QuestsReset_GlobalName:Show()
        
        if TMT_QuestsNamesToggle then -- avoid searching for names if the frame is hidden while in the QuestLog
            TMT_QuestsFrame_GlobalName:Show()
            -- Frame is shown : display "-" button
            TMT_QuestsToggle_GlobalName:SetNormalTexture("Interface/Buttons/UI-MinusButton-Up")
            TMT_QuestsToggle_GlobalName:SetHighlightTexture("Interface/Buttons/UI-MinusButton-Up")
            TMT_QuestsToggle_GlobalName:SetPushedTexture("Interface/Buttons/UI-MinusButton-Down")
            TMT_QuestsFrame_GlobalName:SetHeight((questsLogEntries * TMT_VIncrementsPerChar) + TMT_VMargin * 4)

            local questName
            local toDisplay = ""
            local maxLength = 0
            local currentLength = 0
            local currentPool = {} -- stores current questlog entries (headers and quests)
            local inPoolId
            
            -- Sets the text to show in the frame
            for i = 1, questsLogEntries do
                local title, level, suggestedGroup, isHeader, isCollapsed, isComplete, frequency, questID, startEvent, displayQuestID, isOnMap, hasLocalPOI, isTask, isStory = GetQuestLogTitle(i)
                local tagID, tagName, worldQuestType, rarity, isElite, tradeskillLineIndex, displayTimeLeft = GetQuestTagInfo(questID)

                if not isHeader then
                    questName = TMT_FindNameFromId(questID, TMT.quests)
        
                    if questName then
                        inPoolId = "quest: ".. questID .. "_" .. questName .. "_" .. level

                        -- Adjust the frame length to contain all the text
                        currentLength = TMT_GetStringWidth(questName)
                        if  currentLength > maxLength then
                            maxLength = currentLength
                        end
                        
                        -- Doesn't create a new element if it has already been added to the pool
                        if not TMT_QuestsFramePool[inPoolId] then
                            local questButton = CreateFrame("Button", inPoolId .. "_GlobalName", TMT_QuestsFrame_GlobalName, BackdropTemplateMixin and "BackdropTemplate")
                            questButton:SetNormalFontObject("QuestFontNormalSmall")
                            --questButton:SetPushedTexture("QuestFontNormalSmall")
                            questButton:SetHeight(12.5)
                            questButton:SetScript("OnClick", function(self) TMT_QuestsFrameButtonClick(self, true) end)
                            questButton:SetScript("OnEnter", function(self) TMT_MousePassingByButton(self, TMT_Colors.white) end)
                            questButton:SetScript("OnLeave", function(self) TMT_MousePassingByButton(self, TMT_GetQuestColor(level))  end)
                            questButton:SetText(TMT_GetQuestColor(level) .. questName)
                            questButton:GetFontString():SetPoint("TOPLEFT", questButton, "TOPLEFT", TMT_HMargin * 2, 0)
                                
                            TMT_QuestsFramePool[inPoolId] = questButton
                        else -- allow to change the color without /reload, and remove tag if needed
                            TMT_QuestsFramePool[inPoolId]:SetText(TMT_GetQuestColor(level) .. questName)
                        end

                        table.insert(currentPool, inPoolId)
                    end
                else
                    -- Adjust the frame length to contain all the text
                    currentLength = TMT_GetStringWidth(title)
                    if  currentLength > maxLength then
                        maxLength = currentLength
                    end

                    inPoolId = "header: " .. title
                    if not TMT_QuestsFramePool[inPoolId] then
                        if QuestLogEx or QuestGuru then -- headers buttons can't work with those addons
                            local headerFS = TMT_QuestsFrame_GlobalName:CreateFontString(nil, "ARTWORK", "QuestFontNormalSmall")
                            headerFS:SetText(TMT_Colors.lightGrey .. title)
                            headerFS:SetJustifyH("LEFT");
                            TMT_QuestsFramePool[inPoolId] = headerFS
                        else
                            local headerButton = CreateFrame("Button", inPoolId .. "_GlobalName", TMT_QuestsFrame_GlobalName, BackdropTemplateMixin and "BackdropTemplate")
                            headerButton:SetNormalFontObject("QuestFontNormalSmall")
                            --headerButton:SetPushedTexture("QuestFontNormalSmall")
                            headerButton:SetHeight(12.5)
                            headerButton:SetScript("OnClick", function(self) TMT_QuestsFrameButtonClick(self, false) end)
                            headerButton:SetScript("OnEnter", function(self) TMT_MousePassingByButton(self, TMT_Colors.white) end)
                            headerButton:SetScript("OnLeave", function(self) TMT_MousePassingByButton(self, TMT_Colors.lightGrey)  end)
                            headerButton:SetText(TMT_Colors.lightGrey .. title)
                            headerButton:GetFontString():SetPoint("TOPLEFT")

                            TMT_QuestsFramePool[inPoolId] = headerButton
                        end
                    end

                    table.insert(currentPool, inPoolId) 
                end
            end

            TMT_QuestsLogEntriesPool = currentPool

            if TMT_QuestsNames_DisplayQuestsTags and not QuestGuru and not QuestLogEx then
                maxLength = TMT_SetQuestsTags(maxLength)
            end

            TMT_QuestsFrame_GlobalName:SetWidth((TMT_GetStringWidth(TMT_Tab) + maxLength) + TMT_HMargin * 2)
            TMT_QuestsFramePoolDisplay(TMT_GetStringWidth(TMT_Tab) + maxLength  + TMT_HMargin * 1.5)
            TMT_HighlightEntry(GetQuestLogSelection())
        else 
            TMT_QuestsFrame_GlobalName:Hide()
            -- Frame is hidden : display "+" button
            TMT_QuestsToggle_GlobalName:SetNormalTexture("Interface/Buttons/UI-PlusButton-Up")
            TMT_QuestsToggle_GlobalName:SetHighlightTexture("Interface/Buttons/UI-PlusButton-Up")
            TMT_QuestsToggle_GlobalName:SetPushedTexture("Interface/Buttons/UI-PlusButton-Down")
        end
    else
        -- No quest to translate
        TMT_QuestsFrame_GlobalName:Hide()
        TMT_QuestsToggle_GlobalName:Hide()
        TMT_QuestsReset_GlobalName:Hide()
    end
end

local function TMT_ToggleQuestsFrame()
    TMT_QuestsNamesToggle = not TMT_QuestsNamesToggle
    TMT_FindQuestsOriginalNames()
end

-- Called every time the screen displays another frame, when the questlog is opened
local function TMT_UpdateQuestFrame() 
    local questsLogEntries 

    if QuestGuru then
        -- QuestGuru expands all headers at all times
        -- so there's no update I can do unfortunately
    else
        questsLogEntries = GetNumQuestLogEntries()
    end

    -- Search names only if needed, not every frame
    -- (Except for QuestGuru because I Couldn't figure out how to catch change in entries)
    if questsLogEntries ~= TMT_QuestsLogEntries then
        TMT_FindQuestsOriginalNames()
    end
    
    TMT_QuestsLogEntries = questsLogEntries
end

local function TMT_FindSpellOriginalName(spellId)
    local spellName
    local spellColor

    spellName = TMT_FindNameFromId(spellId, TMT.spells)

    if TMT_SpellTooltip_UseClassColor then
        spellColor = TMT_GetClassColor()
    else
        spellColor = TMT_RgbToHex(TMT_SpellColor)
    end

    return spellName, spellColor
end

local function TMT_FindItemOriginalName(itemLink, tooltip)
    local itemName
    local color

    -- item link of the form "item:7073:0:0:0:0:0:0:0:0:0:0:0:0". 7073 is the ID here
    if itemLink then
        local itemID = tonumber(string.match(itemLink, "item:(%d+)"))
        color = string.match(itemLink, "(|c.+)|H")

        if not TMT_ItemTooltip_UseBlizzardsColor then
            color = TMT_RgbToHex(TMT_ItemColor)
        end

        itemName = TMT_FindNameFromId(itemID, TMT.items)
    end

    return itemName, color
end

local function TMT_GetInstanceInfo()
    local name, instanceType, difficultyID, difficultyName, maxPlayers, dynamicDifficulty, isDynamic, instanceId = GetInstanceInfo()

    return instanceId, name
end

local function TMT_GetMinimapAreaId()
    local currentAreaName = GetMinimapZoneText()
    local currentAreaId
    local areaTable

    for i = 1, TMT_AreasMaxId do
        local areaName = C_Map.GetAreaInfo(i)

        if areaName == currentAreaName then
            currentAreaId = i
            areaTable = TMT.areas
            break;
        end
    end

    return currentAreaId
end

local function TMT_FindAreaOriginalName(areaId)
    local areaName = TMT_FindNameFromId(areaId, TMT.areas)

    return areaName
end

local function TMT_GetAreaNameColor()
    return "|r"
end

local function TMT_EntryPoint(isOn, task, param1, param2)
    if isOn then
        if task == "quests" then
            TMT_FindQuestsOriginalNames()
            if GetNumQuestLogEntries() > 0 then
                TMT_HighlightEntry(GetQuestLogSelection())
            end
        elseif task == "spell" then
            -- param1 : spell id, param2 : toolTip            
            local name, color = TMT_FindSpellOriginalName(param1)
            TMT_TooltipDisplay(param2, name, TMT_SpellTooltip_DisplayNameInTitle, color)              
        elseif task == "talent" then
            local color 
            -- param1: talent name, param2 :tooltip
            if TMT_SpellTooltip_UseClassColor then
                color = TMT_GetClassColor()
            else
                color = TMT_RgbToHex(TMT_SpellColor)
            end

            TMT_TooltipDisplay(param2, param1, TMT_SpellTooltip_DisplayNameInTitle, color)
        elseif task == "npc" then
            local name, color = TMT_FindNpcOriginalName(param1)
            TMT_TooltipDisplay(param2, name, TMT_NPCTooltip_DisplayNameInTitle, color)
        elseif task == "updateQuests" then
            TMT_UpdateQuestFrame()
        elseif task == "toggleQuests" then
            TMT_ToggleQuestsFrame()
        elseif task == "highlightEntry" then
            if GetNumQuestLogEntries() > 0 then
                TMT_HighlightEntry(GetQuestLogSelection())
            end
        elseif task == "item" then
            -- param1 : item link, param2 : toolTip
            local name, color = TMT_FindItemOriginalName(param1, param2)

            if not TMT_IsNameAlreadyInTooltip(param2, name) then -- script called twice for recipes
                TMT_TooltipDisplay(param2, name, TMT_ItemTooltip_DisplayNameInTitle, color)
            end
        elseif task == "area" then
            
            --param1 : tooltip
            local areaId = TMT_GetMinimapAreaId()

            if areaId then
                local name = TMT_FindAreaOriginalName(areaId)
                local color = TMT_GetAreaNameColor()
                
                -- Sometimes the English name is also given with the localized name by Blizzard
                if not string.match(GetMinimapZoneText(), name) then
                    TMT_TooltipDisplay(param1, name, TMT_MinimapAreaTooltip_DisplayNameInTitle, color)
                    GameTooltip:Show()
                end
            end
        end
    else
        if task == "quests" or task == "updateQuests" then
            TMT_QuestsFrame_GlobalName:Hide()
            TMT_QuestsToggle_GlobalName:Hide()
            TMT_QuestsReset_GlobalName:Hide()
        end
    end
end

--[[
    A few words on setUnitAura, setUnitBuff, etc.
    - setUnitAura is to work on the buffs AND debuffs on YOUR character
    - setUnitBuff is to work ONLY on the buffs on your target (can be yourself)
    - setUnitDebuff is to work ONLY on the debuffs on your target (can be yourself)
    - UnitAura retrieves ONLY buffs on your character and target
    - UnitBuff retrieves buffs on you and your target
    - UnitDebuff retrieves debuffs on you and your target

    Edit :
    My understanding for setUnitAura is only right on default Blizzard behavior. But
    it seems that it can be called on target with other addons, like Shadowed Unit Frame.
]]

-- Securehook handler. Called only on player with Blizzard. Called also on target with ShadowedUF and LunaUF
local function TMT_FindAuraID(casterID, unitID, auraIndex, filter)
        local actualSpellId

        -- The NOT_PASSIVE version was added in patch 1.13.3 for buffs and debuffs on yourself (but not yourself from target)
        if filter == "HELPFUL" or filter == "HELPFUL_NOT_PASSIVE" then
            local name, rank, count, debuffType, duration, expirationTime, unitCaster, isStealable, 
            shouldConsolidate, spellId = UnitBuff(unitID, auraIndex)
            
            actualSpellId = spellId
        elseif filter == "HARMFUL" or filter == "HARMFUL_NOT_PASSIVE" then
            local name, rank, count, debuffType, duration, expirationTime, unitCaster, isStealable, 
            shouldConsolidate, spellId = UnitDebuff(unitID, auraIndex)
    
            actualSpellId = spellId
        end
    
        TMT_EntryPoint(TMT_DisplaySpellsNames, "spell", actualSpellId, GameTooltip)
         -- since it's a secure hook, GameTooltip was already set. Have to show it again to adapt size to new text
        GameTooltip:Show()
end

-- Securehook handler
local function TMT_FindTargetBuffID(casterID, unitID, buffIndex)
    local name, rank, count, debuffType, duration, expirationTime, unitCaster, isStealable, 
    shouldConsolidate, spellId = UnitAura("TARGET", buffIndex)
        
    TMT_EntryPoint(TMT_DisplaySpellsNames, "spell", spellId, GameTooltip)
    GameTooltip:Show()
end

-- Securehook handler
local function TMT_FindTargetDebuffID(casterID, unitID, buffIndex)
    local name, rank, count, debuffType, duration, expirationTime, unitCaster, isStealable, 
    shouldConsolidate, spellId = UnitDebuff("TARGET", buffIndex)
  
    TMT_EntryPoint(TMT_DisplaySpellsNames, "spell", spellId, GameTooltip)
    GameTooltip:Show()
end

local function TMT_TalentTooltipSecureHook(self, tabIndex, talentIndex)
    local localeClass, englishClass, classIndex = UnitClass("player")
    local talentName = TMT.talents[classIndex][tabIndex][talentIndex]
    local tooltipLine = ""

    -- If the talent is also a learnable book spell, setSpell is called on tooltip before setTalent. We don't want to display the name twice
    if TMT_SpellTooltip_DisplayNameInTitle then
        -- The english name is on the first line, after "- ", or on a new line
        local placement = "- (.+)"
        if TMT_Tooltips_DisplayNameUnderTitle then 
            placement = "|n(.+)"
        end
        
        tooltipLine = string.match(GameTooltipTextLeft1:GetText(), placement)
    else
        tooltipLine = _G["GameTooltipTextLeft" .. GameTooltip:NumLines()]:GetText()
    end

    if tooltipLine then
        tooltipLine = string.sub(tooltipLine, 11) -- removes the color
    end

    if talentName ~= tooltipLine then
        TMT_EntryPoint(TMT_DisplaySpellsNames, "talent", talentName, self)
        self:Show()
    end
end

-- Useful for quests frame support in other addons. Changes the frame hosting TMT quests frame
local function TMT_AddonsSupport()
    if QuestLogEx then -- QuestLogEx support
        TMT_QuestsFrameParent = QuestLogExFrame
        TMT_Defaults.questsFrameX  = -38
        TMT_Defaults.questsFrameY = -70
    elseif QuestGuru then  -- QuestGuru support
        TMT_QuestsFrameParent = QuestGuru
        TMT_Defaults.questsFrameX = -6
        TMT_Defaults.questsFrameY = -70
    elseif LightHeadedFrame then
        TMT_QuestsFrameParent = QuestLogFrame
        TMT_Defaults.questsFrameX = LightHeadedFrame:GetWidth() - 53
        TMT_Defaults.questsFrameY = -70
    else -- No dependency
        TMT_QuestsFrameParent = QuestLogFrame
        TMT_Defaults.questsFrameX  = -39.5
        TMT_Defaults.questsFrameY = -70
    end

    if not TMT_QuestsFrame_PositionChangedByUser then
        TMT_QuestsFrameX = TMT_Defaults.questsFrameX
        TMT_QuestsFrameY = TMT_Defaults.questsFrameY
    end

    TMT_QuestsFrameToggleButtonX = TMT_Defaults.questsFrameX + 2
    TMT_QuestsFrameToggleButtonY = TMT_Defaults.questsFrameY + 10
end

local function TMT_StopMoving()
    _G["TMT_QuestsFrame_GlobalName"]:StopMovingOrSizing()
    TMT_QuestsFrame_PositionChangedByUser = true
    
    point, relativeTo, relativePoint, x, y = _G["TMT_QuestsFrame_GlobalName"]:GetPoint()

    -- relativeTo always nil
    TMT_QuestsFramePoint = point
    TMT_QuestsFrameRelativePoint = relativePoint
    TMT_QuestsFrameX = x
    TMT_QuestsFrameY = y
end

-- Frame to display the quests in English
local function TMT_SetQuestsFrame()
    TMT_QuestsFrame = CreateFrame("Frame", "TMT_QuestsFrame_GlobalName", TMT_QuestsFrameParent, BackdropTemplateMixin and "BackdropTemplate")

    -- Makes the frame draggable
    TMT_QuestsFrame:SetMovable(true)
    TMT_QuestsFrame:EnableMouse(true)
    TMT_QuestsFrame:RegisterForDrag("LeftButton")
    TMT_QuestsFrame:SetScript("OnDragStart", TMT_QuestsFrame.StartMoving)
    TMT_QuestsFrame:SetScript("OnDragStop", TMT_StopMoving)

    -- if user moved the frame, the position is relatively to the screen, not the parent frame
    if not TMT_QuestsFrame_PositionChangedByUser then
        TMT_QuestsFrame:SetPoint(TMT_QuestsFramePoint, TMT_QuestsFrameParent, TMT_QuestsFrameRelativePoint, TMT_QuestsFrameX, TMT_QuestsFrameY)
    else
        TMT_QuestsFrame:SetPoint(TMT_QuestsFramePoint, UIParent, TMT_QuestsFrameRelativePoint, TMT_QuestsFrameX, TMT_QuestsFrameY)
    end
        
    TMT_QuestsFrame:SetBackdrop({
        bgFile = "Interface/Stationery/StationeryTest1",
        edgeFile ="Interface/DialogFrame/UI-DialogBox-Border",
        tile = false, 
        tileSize = 0,
        edgeSize = 20,
        insets = {left = 4, right = 4, top = 4, bottom = 4 }
    });
    TMT_QuestsFrame:SetBackdropColor(0.1, 0.1, 0.1, 0.985)

    -- Button to show/hide quests names in english
    TMT_QuestsToggle = CreateFrame("Button", "TMT_QuestsToggle_GlobalName", TMT_QuestsFrameParent)
    TMT_QuestsToggle:SetWidth(16)
    TMT_QuestsToggle:SetHeight(16)
    TMT_QuestsToggle:SetPoint("BOTTOMLEFT", TMT_QuestsFrameParent, "TOPRIGHT", TMT_QuestsFrameToggleButtonX, TMT_QuestsFrameToggleButtonY)

    -- Button to put the quests frame back to the default position
    TMT_QuestsReset = TMT_DefaultButton(TMT_QuestsFrame, "TOPRIGHT", "BOTTOMRIGHT", 70, 16, 0, -4, "Hook back", "GameFontNormalSmall", "TMT_QuestsReset_GlobalName")
    TMT_QuestsReset:SetClampedToScreen(true)
    TMT_QuestsReset:SetScript("OnClick", function(self) TMT_RestoreDefaults_QuestsFramePosition() end)
end

local function TMT_SetHooks()
    -- Hooks to tooltip events
    GameTooltip:HookScript("OnTooltipSetUnit", function(self)  TMT_EntryPoint(TMT_DisplayNPCsNames, "npc", select(2, self:GetUnit()), self); end)
    GameTooltip:HookScript("OnTooltipSetSpell", function(self) TMT_EntryPoint(TMT_DisplaySpellsNames, "spell", select(2, self:GetSpell()), self); end)
    GameTooltip:HookScript("OnTooltipSetItem", function(self) TMT_EntryPoint(TMT_DisplayItemsNames, "item", select(2, self:GetItem()), self); end)
    ItemRefTooltip:HookScript("OnTooltipSetItem", function(self) TMT_EntryPoint(TMT_DisplayItemsNames, "item", select(2, self:GetItem()), self); end) -- tooltip when an item link is clicked
    ShoppingTooltip1:HookScript("OnTooltipSetItem", function(self) TMT_EntryPoint(TMT_DisplayItemsNames, "item", select(2, self:GetItem()), self); end) -- tooltip when comparing items
    ShoppingTooltip2:HookScript("OnTooltipSetItem", function(self) TMT_EntryPoint(TMT_DisplayItemsNames, "item", select(2, self:GetItem()), self); end) 
    ItemRefShoppingTooltip1:HookScript("OnTooltipSetItem", function(self) TMT_EntryPoint(TMT_DisplayItemsNames, "item", select(2, self:GetItem()), self); end) -- tooltip when comparing items but from an item link
    ItemRefShoppingTooltip2:HookScript("OnTooltipSetItem", function(self) TMT_EntryPoint(TMT_DisplayItemsNames, "item", select(2, self:GetItem()), self); end)
    hooksecurefunc(GameTooltip, "SetUnitAura", TMT_FindAuraID) -- Called on your character
    hooksecurefunc(GameTooltip, "SetUnitBuff", TMT_FindTargetBuffID) -- Only called on target (can be yourself)
    hooksecurefunc(GameTooltip, "SetUnitDebuff", TMT_FindTargetDebuffID) -- Only called on target (can be yourself)
    hooksecurefunc(GameTooltip, "SetTalent", TMT_TalentTooltipSecureHook)

    -- hooks to questlog events
    TMT_QuestsFrameParent:HookScript("OnShow", function(self)  TMT_EntryPoint(TMT_DisplayQuestsNames, "quests"); end)
    TMT_QuestsToggle_GlobalName:HookScript("OnClick", function(self) TMT_EntryPoint(TMT_DisplayQuestsNames, "toggleQuests"); end)
    TMT_QuestsFrameParent:HookScript("OnUpdate", function(self)  TMT_EntryPoint(TMT_DisplayQuestsNames, "updateQuests"); end)
    _G["QuestLogTitle1"]:HookScript("OnClick", function(self) TMT_EntryPoint(TMT_DisplayQuestsNames, "highlightEntry") end)
    _G["QuestLogTitle2"]:HookScript("OnClick", function(self) TMT_EntryPoint(TMT_DisplayQuestsNames, "highlightEntry") end)
    _G["QuestLogTitle3"]:HookScript("OnClick", function(self) TMT_EntryPoint(TMT_DisplayQuestsNames, "highlightEntry") end)
    _G["QuestLogTitle4"]:HookScript("OnClick", function(self) TMT_EntryPoint(TMT_DisplayQuestsNames, "highlightEntry") end)
    _G["QuestLogTitle5"]:HookScript("OnClick", function(self) TMT_EntryPoint(TMT_DisplayQuestsNames, "highlightEntry") end)
    _G["QuestLogTitle6"]:HookScript("OnClick", function(self) TMT_EntryPoint(TMT_DisplayQuestsNames, "highlightEntry") end)

    -- hooks related to the current arey
    MinimapZoneTextButton:HookScript("OnEnter", function(self) TMT_EntryPoint(TMT_DisplayAreasNames, "area", GameTooltip); end)

    -- special hooks for addons support
    if QuestLogEx then 
        QuestLogExFrameDescription:HookScript("OnShow", 
            function(self) 
                if not TMT_QuestsFrame_PositionChangedByUser then
                    TMT_ChangeFramePosition(TMT_QuestsFrame_GlobalName, "TOPLEFT", QuestLogExFrameDescription, "TOPRIGHT", -29, TMT_QuestsFrameToggleButtonY);
                end

                TMT_ChangeFramePosition(TMT_QuestsToggle_GlobalName, "TOPLEFT", QuestLogExFrameDescription, "TOPRIGHT", -26, TMT_QuestsFrameY + 10)
            end
        )

        QuestLogExFrameDescription:HookScript("OnHide", 
            function(self) 
                if not TMT_QuestsFrame_PositionChangedByUser then
                    TMT_ChangeFramePosition(TMT_QuestsFrame_GlobalName, "TOPLEFT", QuestLogExFrame, "TOPRIGHT", TMT_QuestsFrameX, TMT_QuestsFrameToggleButtonY);
                end

                TMT_ChangeFramePosition(TMT_QuestsToggle_GlobalName, "TOPLEFT", QuestLogExFrame, "TOPRIGHT", TMT_QuestsFrameToggleButtonX + 4, TMT_QuestsFrameToggleButtonY + 10)
            end
        )
    end
end

-- Need this frame for saved variables
local TMT_eventFrame = CreateFrame("FRAME"); -- Needed to respond to events
TMT_eventFrame:RegisterEvent("ADDON_LOADED"); -- Fired when saved variables loaded
TMT_eventFrame:RegisterEvent("PLAYER_LOGIN"); 
TMT_eventFrame:RegisterEvent("PLAYER_LOGOUT"); -- Fired when player about to log out

function TMT_eventFrame:OnEvent(event, addonName)
    if event == "ADDON_LOADED" and addonName == "TheMotherTongue" then
        if TMT_DisplayNPCsNames == nil then
            print(TMT_Colors.orange .. "Type \"/tmt\" to go to TheMotherTongue addon settings")
            TMT_DisplayNPCsNames = TMT_Defaults.displayNPCsNames
        end
        if TMT_NPCTooltip_DisplayNameInTitle == nil then -- first time addon is launched
            TMT_NPCTooltip_DisplayNameInTitle = TMT_Defaults.npcTooltip_displayNameInTitle
        end
        if TMT_NamesDump == nil then
            TMT_NamesDump = {}
        end
        if TMT_DisplayQuestsNames == nil then 
            TMT_DisplayQuestsNames = TMT_Defaults.displayQuestsNames
        end
        if TMT_QuestsNamesToggle == nil then 
            TMT_QuestsNamesToggle = TMT_Defaults.questNamesToggle
        end
        if TMT_DisplaySpellsNames == nil then 
            TMT_DisplaySpellsNames = TMT_Defaults.displaySpellsNames
        end
        if TMT_SpellTooltip_DisplayNameInTitle == nil then 
            TMT_SpellTooltip_DisplayNameInTitle = TMT_Defaults.spellTooltip_displayNameInTitle
        end
        if TMT_NpcColor == nil then
            TMT_NpcColor = TMT_Defaults.npcColor
        end
        if TMT_SpellColor == nil then
            TMT_SpellColor = TMT_Defaults.spellColor
        end
        if TMT_QuestColor == nil then
            TMT_QuestColor = TMT_Defaults.questColor
        end
        if TMT_QuestTooltip == nil then
            TMT_QuestTooltip = TMT_Defaults.questTooltip
        end
        if TMT_QuestsFrame_HighlightBackdropColor == nil then
            TMT_QuestsFrame_HighlightBackdropColor = TMT_Defaults.questsFrame_HighlightBackdropColor
        end
        if TMT_DisplayItemsNames == nil then
            TMT_DisplayItemsNames = TMT_Defaults.displayItemsNames
        end
        if TMT_ItemTooltip_DisplayNameInTitle == nil then
            TMT_ItemTooltip_DisplayNameInTitle = TMT_Defaults.itemTooltip_displayNameInTitle
        end
        if TMT_ItemColor == nil then
            TMT_ItemColor = TMT_Defaults.itemColor
        end
        if TMT_ItemTooltip_UseBlizzardsColor == nil then
            TMT_ItemTooltip_UseBlizzardsColor = TMT_Defaults.itemTooltip_UseBlizzardsColor
        end
        if TMT_QuestsNames_UseBlizzardsColor == nil then
            TMT_QuestsNames_UseBlizzardsColor = TMT_Defaults.questsNames_UseBlizzardsColor
        end
        if TMT_QuestsNames_DisplayQuestsTags == nil then
            TMT_QuestsNames_DisplayQuestsTags = TMT_Defaults.questsNames_DisplayQuestsTags
        end
        if TMT_QuestsFrameX == nil then
            TMT_QuestsFrameX = TMT_Defaults.questsFrameX
        end
        if TMT_QuestsFrameY == nil then
            TMT_QuestsFrameY = TMT_Defaults.questsFrameY
        end
        if TMT_QuestsFrame_PositionChangedByUser == nil then
            TMT_QuestsFrame_PositionChangedByUser = TMT_Defaults.questsFrame_PositionChangedByUser
        end
        if TMT_QuestsFramePoint == nil then
            TMT_QuestsFramePoint = TMT_Defaults.questsFramePoint
        end
        if TMT_QuestsFrameRelativePoint == nil then
            TMT_QuestsFrameRelativePoint = TMT_Defaults.questsFrameRelativePoint
        end
        if TMT_SpellTooltip_UseClassColor == nil then
            TMT_SpellTooltip_UseClassColor = TMT_Defaults.spellTooltip_UseClassColor
        end
        if TMT_Tooltips_DisplayNameUnderTitle == nil then
            TMT_Tooltips_DisplayNameUnderTitle = TMT_Defaults.tooltips_DisplayNameUnderTitle
        end
        if TMT_DisplayAreasNames == nil then
            TMT_DisplayAreasNames = TMT_Defaults.displayAreasNames
        end
        if TMT_MinimapAreaTooltip_DisplayNameInTitle == nil then
            TMT_MinimapAreaTooltip_DisplayNameInTitle = TMT_Defaults.areaTooltip_DisplayNameInTitle
        end
        if TMT_Language == nil then
            TMT_Language = TMT_Defaults.language
        end
        if TMT_LastUpdateMessageVersion == nil then
            TMT_LastUpdateMessageVersion = TMT_Defaults.lastUpdateMessageVersion
        end
    elseif event == "PLAYER_LOGIN" then
        TMT_QuestsLogEntries = GetNumQuestLogEntries()
       -- Tests the existence of languages packs
       TMT_DetectLanguagesPacks()

        --[[
            We need to create the panel on login if we want to use saved variables to
            set the options (to match user preferences)
        ]]
        TMT_CreateConfigPanel()

        -- Doing this now allows for hooking other addons events, once they are all loaded
        TMT_AddonsSupport()
        TMT_SetQuestsFrame()
        TMT_SetHooks()
        TMT_DisplayUpdateMessage()
    end
end
TMT_eventFrame:SetScript("OnEvent", TMT_eventFrame.OnEvent);